/*
 * iFunc.cpp
 *
 *  Created on: 18 окт. 2016 г.
 *      Author: ryutin
 */

#include <cmath>
#include <complex>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include <utility>
#include <vector>
#include <stdio.h>
#include <cassert>

#include "../inc/I.h"
#include "../inc/PI.h"
#include "../inc/NDarray.h"
#include "../inc/NDgrid.h"
#include "../inc/iFunc.h"


///////////////////////////////////////////////////////////////////
ExDiff::iFunc::iFunc (const ExDiff::NDarray<double>& _iTab, const ExDiff::NDgrid<double>& _iGrid):
	    iTab(_iTab),
		iGrid(_iGrid),
		dimensions(CalcDim(_iTab)),
		vbin(CalcBin(_iTab)),
		vbinsize(1),		
		minvars(_iGrid.GetValues()),
		maxvars(_iGrid.GetValues()),
		point(_iGrid.GetValues()),
		cubeindex(CalcDim(_iTab)),
		cubemin(_iGrid.GetValues()),
		cubemax(_iGrid.GetValues()),
		maxfunindexes(CalcDim(_iTab)),
		minfunindexes(CalcDim(_iTab)),	
	    maxfunvalue(1.0),
	    maxfunpoint(_iGrid.GetValues()),
	    minfunvalue(0.0),
	    minfunpoint(_iGrid.GetValues())
{
              assert(!iTab.GetValues().empty());
              assert(!iGrid.GetValues().empty());
			  assert(!vbin.GetValues().empty());
			  assert(vbinsize > 0);
			  assert(!dimensions.empty());           
              assert(!minvars.empty());
              assert(!maxvars.empty());
              assert(!point.empty());
              assert(!cubeindex.empty());
              assert(!cubemin.empty());
              assert(!cubemax.empty());			  
              assert(!maxfunindexes.empty());			  
              assert(!minfunindexes.empty());				  
              assert(maxfunvalue > 0.0);
              assert(!maxfunpoint.empty());
              assert(minfunvalue >= 0.0 && minfunvalue <= maxfunvalue);
              assert(!minfunpoint.empty());

// initialization with real values
			  vbin = CalcBin(_iTab);
              vbinsize = vbin.computeTotalSize(vbin.GetDimensions());	  
			  
			  minvars.clear();
			  maxvars.clear();
			  point.clear();
			  cubemin.clear();
			  cubemax.clear();
			  maxfunindexes.clear();
			  minfunindexes.clear();
			  maxfunpoint.clear();
			  minfunpoint.clear();
			  			  
			  minvars = CalcMinVars(_iGrid);		
       		  maxvars = CalcMaxVars(_iGrid);
		      point = CalcMinVars(_iGrid);		
		      cubemin = CalcMinVars(_iGrid);
		      cubemax = CalcMaxVars(_iGrid);	
			  
              std::vector<double> v = _iTab.GetValues();
              std::size_t nmax = CalcMax(_iTab);
			  std::size_t nmin = CalcMin(_iTab);
			  
              maxfunindexes = _iTab.computeIndexes(nmax);
              minfunindexes = _iTab.computeIndexes(nmin);			  
			  maxfunvalue = v[nmax];
			  maxfunpoint = iGrid.get(maxfunindexes);
			  minfunvalue = v[nmin];
			  minfunpoint = iGrid.get(minfunindexes);        		  	      			  
};
//
ExDiff::iFunc::~iFunc(){
	
};
// print information
void ExDiff::iFunc::printC(){
	std::cout<< "iFunc printC() ============================================= " <<std::endl;	
	iTab.printC();
	iGrid.printC();
	iGrid.printGP();
    vbin.printC();	
	std::cout<< "iFunc values ==================================== " <<std::endl;	
	std::size_t n = dimensions.size();
	std::cout<< "minvars = " <<std::endl;
	PrintVector(minvars,n);
	std::cout<< "maxvars = " <<std::endl;
	PrintVector(maxvars,n);
	std::cout<< "point = " <<std::endl;
	PrintVector(point,n);
	std::cout<< "cubeindex = " <<std::endl;
	PrintVector(cubeindex,n);
	std::cout<< "cubemin = " <<std::endl;
	PrintVector(cubemin,n);
	std::cout<< "cubemax = " <<std::endl;
	PrintVector(cubemax,n);
    std::cout<< "MAXFUN = " <<  maxfunvalue <<std::endl;	
    PrintVector(maxfunindexes,n);	
	PrintVector(maxfunpoint,n);
    std::cout<< "MINFUN = " <<  minfunvalue <<std::endl;	
    PrintVector(minfunindexes,n);	
	PrintVector(minfunpoint,n);		
    std::cout<<"============================================================== " <<std::endl;
};


// constant binary NDarray calculation
ExDiff::NDarray<size_t> ExDiff::iFunc::CalcBin(const ExDiff::NDarray<double>& _iTab){
    std::vector<size_t> vdim = _iTab.GetDimensions();
	std::vector<size_t> vindexes = vdim;
	std::size_t vdimsize = 1;	
	for(std::size_t i = 0; i < dimensions.size(); i++){
		vdimsize *=2;
	    vdim[i] = 2;
	} 	
	ExDiff::NDarray<size_t> vbin_(vdim);
	for(std::size_t i = 0; i < vdimsize; i++){
	  vindexes = vbin_.computeIndexes(i);
	  vbin_.get(vindexes) = i;
	}
	return vbin_;
};

// check that point is inside the range
int ExDiff::iFunc::CheckPoint(std::vector<double>& _point){
	int n=0;
	for(std::size_t i = 0; i < _point.size(); i++){
	  if(_point[i] < minvars[i] || _point[i] > maxvars[i] ){
		  n=1;
	  }	
	}
//	std::cout<< "n_checkpoint = " << n <<std::endl; 
	return n;
};	

// aux function FunDif

double ExDiff::iFunc::FunDif(std::size_t& n_, std::size_t& varindex_, std::vector<double>& _point){
	if(n_==0){
		return (cubemax[varindex_] - _point[varindex_])/(cubemax[varindex_]-cubemin[varindex_]);
	} else{
	    return (_point[varindex_] - cubemin[varindex_])/(cubemax[varindex_]-cubemin[varindex_]);
	}
};

// Calc: calculation of the interpolation function at the _point	

double ExDiff::iFunc::Calc(std::vector<double> _point){
	assert(_point.size() == dimensions.size());
	CalcCube(_point);
	
	point = _point;

	double value = 0.0;
	double valuef = 1.0;
	std::vector<size_t> v = dimensions;
	std::vector<size_t> v1 = dimensions;
    if(CheckPoint(_point)==0){
// use cube variables to calculate the value			
	  for(std::size_t i = 0; i < vbinsize; i++){
       v = vbin.computeIndexes(i);

	    for(std::size_t j = 0; j < v.size(); j++){
          v1[j] = cubeindex[j]+v[j];
          valuef *= FunDif(v[j],j,_point); 		 	 		  
	    }

	   valuef *= iTab.get(v1);

       value += valuef; 
       valuef = 1.0; 	   
	  }			
	  return value;	
	} else{
	    std::cout<< "Point is out of the interpolation range! Calculation is stopped!" <<std::endl;	
	    return value;	
	}
};

/////////////////////////////////////////////////////////////////////////////
// extraction of minimal values of variables
std::vector<double> ExDiff::iFunc::CalcMinVars(const ExDiff::NDgrid<double>& _iGrid){
	double r=0.0;
	for(std::size_t i=0; i < dimensions.size(); i++){
		r = _iGrid.GetVar(i,0);
		minvars.push_back(r);
	}
	return minvars;
};
// extraction of maximal values of variables
std::vector<double> ExDiff::iFunc::CalcMaxVars(const ExDiff::NDgrid<double>& _iGrid){
	double r=0.0;
	for(std::size_t i=0; i < dimensions.size(); i++){
		r = _iGrid.GetVar(i,dimensions[i]-1);
		maxvars.push_back(r);
	}	
	return maxvars;
};
// calculation of the hyper-cube for the interpolation from  the point
void ExDiff::iFunc::CalcCube(std::vector<double>& _point){
// use iGrid.GetVar(i,j) and _point[i], cubeindex[i]  
    std::size_t j=0; 
    std::vector<size_t> cubeindexmax = cubeindex;

    if(CheckPoint(_point)==0){		
      for(std::size_t i=0; i < _point.size(); i++){
		  j = 0;
		  while(_point[i] >= iGrid.GetVar(i,j) && j < dimensions[i]-1){
		   j++;	
		  }
		  cubeindex[i]=j-1;
		  cubeindexmax[i]=j;
	  } 
      cubemin = iGrid.get(cubeindex);
	  cubemax = iGrid.get(cubeindexmax);
   } else{
	  std::cout<< "Point is out of the interpolation range! Calculation is stopped!" <<std::endl;	
   }	
   return;
};
//////////////////////////////////////////////////////////////////
// calculation of max fun general index
std::size_t ExDiff::iFunc::CalcMax(const ExDiff::NDarray<double>& _iTab){
    std::vector<double> v = _iTab.GetValues();
	std::size_t n = 0;
    for(std::size_t i=0; i < v.size(); i++){
		if(v[i]>v[n]){
			n=i;
		}
	}
	return n;
};
// calculation of min fun general index
std::size_t ExDiff::iFunc::CalcMin(const ExDiff::NDarray<double>& _iTab){
    std::vector<double> v = _iTab.GetValues();
	std::size_t n = 0;
    for(std::size_t i=0; i < v.size(); i++){
		if(v[i]<v[n]){
			n=i;
		}
	}	
	return n;	
};	
// print vector of points and indexes [0]...[dimensions.size()]
void ExDiff::iFunc::PrintVector(std::vector<size_t> vec_, std::size_t size_){		  	  
	  std::cout << " { ";
          for (std::size_t j = 0; j < size_; j++) {
             std::cout << vec_[j];
			 if(j<size_-1) {
				 std::cout <<  " , ";
			 }
          }	  
	  std::cout << " } " << std::endl;	  		
      return;
};	
void ExDiff::iFunc::PrintVector(std::vector<double> vec_, std::size_t size_){		  	  
	  std::cout << " { ";
          for (std::size_t j = 0; j < size_; j++) {
             std::cout << vec_[j];
			 if(j<size_-1) {
				 std::cout <<  " , ";
			 }
          }	  
	  std::cout << " } " << std::endl;	  		
      return;
};
// aux reduce 1 last dimension
ExDiff::NDarray<double> ExDiff::iFunc::ReduceNDarray(ExDiff::NDarray<double>& Tab_){
	  std::vector<size_t> dim_ = Tab_.GetDimensions();
	  dim_.pop_back();
      ExDiff::NDarray<double> _Tab(dim_);
      return _Tab;		
};	
ExDiff::NDgrid<double> 	ExDiff::iFunc::ReduceNDgrid(ExDiff::NDgrid<double>& Grid_){
	  std::vector<size_t> dim_ = Grid_.GetDimensions();
	  dim_.pop_back();	
      ExDiff::NDgrid<double> _Grid(dim_);
      return _Grid;	
};